// module.cc -- squeakCompiler module and named primitives
//
// Author: Ian.Piumarta@inria.fr
//
// last edited: Thu Feb 25 14:18:05 1999 by piumarta (Ian Piumarta) on pingu

#include "hooks.h"
#include "sqVirtualMachine.h"

#include <stdio.h>


static VirtualMachine *vm;


// these can be called as NAMED PRIMITIVES from Smalltalk (<grin> :^)

// Note: calling compilerDisable will turn off *all* of the Interpreter's
// hooks.  Any structures within the compiler that contain real oops will
// *not* be given the chance to remap during GC.  Re-enabling after a
// disable must therefore perform a "cold start" of the compiler.  (This is
// almost certainly what someone calling compilerDisable wants anyway.)

int squeakCompilerEnable(void)
{
  j_initialiseHook();	// init compiler structures, etc...
  vm->setCompilerInitialized(1);
  return 0;
}


int squeakCompilerDisable(void)
{
  j_processChangeHook();
  vm->setCompilerInitialized(0);
  j_releaseHook();	// release compiler resources, etc...
  return 0;
}


// module initialisation function


int setInterpreter(VirtualMachine *squeak)
{
  vm= squeak;

  if ((vm->majorVersion() != HOOK_MAJOR) ||
      (vm->minorVersion()  < HOOK_MINOR)) {
#   ifdef unix /* good for Unix, bad for elsewhere... */
    fprintf(stderr, "squeakCompiler cannot use VM proxy version %d.%d\n",
	    vm->majorVersion(), vm->minorVersion());
    fprintf(stderr, "requires VM proxy version: major == %d, minor >= %d\n",
	    HOOK_MAJOR, HOOK_MINOR);
#   endif
    return 0;
  }

  CompilerHook *hooks= vm->compilerHookVector();

  // these must match Interpreter methodsFor: 'compiler support'
  hooks[1]= (CompilerHook)j_activateMethodHook;
  hooks[2]= (CompilerHook)j_flushCacheHook;
  hooks[3]= (CompilerHook)j_preGCHook;
  hooks[4]= (CompilerHook)j_mapHook;
  hooks[5]= (CompilerHook)j_postGCHook;
  hooks[6]= (CompilerHook)j_processChangeHook;
  hooks[7]= (CompilerHook)j_preSnapshotHook;
  hooks[8]= (CompilerHook)j_postSnapshotHook;
  hooks[9]= (CompilerHook)j_markHook;

  squeakCompilerEnable();

  return 1;
}
